.srfim.character <- function(dirPath, pattern=character(0), ...)
  {
    ints <- readIntensities(dirPath, pattern=pattern, withVariability=FALSE, ...)
    .srfim.SolexaIntensity(ints, ...)
  }

.srfim.SRATable <- function(dirPath, lane, tile, cycles=NULL, ...)
  {
    tab <- dirPath
    ints <- SRATableReadIntensities(tab, lane=lane, tile=tile, ...)
    if (!missing(cycles) || is.null(cycles)) {
      ints <- ints[,,cycles]
    }
    .srfim.SolexaIntensity(ints, ...)
  }

.srfim.SolexaIntensity <- function(dirPath,
                                   filename="",
                                   filter=c("Entropy","SNR","SNR2","None"),
                                   preprocess=TRUE, ...)
  {
    ints <- dirPath

    if (preprocess) {
      pre.res <- srfimPreprocess(ints)
      model <- srfimModel(pre.res, readInfo(ints))
    } else {
      model <- srfimModel(ints, readInfo(ints))
    }

    qa.res <- srfimQA(model)
    filter <- match.arg(filter)

    if (filter == "Entropy") {
      f <- srfimENTfilter(model)
    } else if (filter == "None") {
      f <- logical(0)
    } else {
      warning("Filter ", filter, " not implemented yet. Ask the maintainer to implement it!")
      warning("Continuing without filtering")
      f <- logical(0)
    }

    rq <- srfimBasecall(model, readInfo(ints), filter=f, filename=filename)
    call <- match.call()
    new("SrfimResult",
        basecalls=rq,
        qa=qa.res,
        filter.type=filter,
        filter=f,
        filename=filename,
        call=call)
  }

.srfimGetFiles.SolexaIntensity <- function(path,
                                           pattern=character(0),
                                           intExtension="_int.txt", ...)
  {
    intPattern <- paste(pattern, intExtension, sep="")
    fls <- ShortRead:::.file_names(path, intPattern)
    return(basename(fls))
  }

.srfim.SrfimBatch.SRATable <- function(object, destdir="", lane=NULL, tiles=NULL, fragments=NULL, ...)
  {
    tab <- SRATable(object@path)
    info <- SRAGetTableInfo(tab)

    if (!is.null(lane)) {
      if (length(lane) != 1) {
        stop("Only one lane at a time!")
      }

      if (! lane %in% info$lanes) {
        stop("Requested lane ", lane, " not found in SRA table")
      }
    } else {
      lane <- info$lanes[1]
    }

    m <- match(lane, info$lanes)
    if (is.null(tiles)) {
      tiles <- 1:info$ntiles[m]
    } else {
      if (min(tiles) < 1 || max(tiles) > info$ntiles[m]) {
        stop("Requested tiles ", tiles, " are out of range")
      }
    }
    
    if (missing(destdir) || nchar(destdir) == 0) {
      destdir <- getwd()
    }

    if (!is.null(fragments)) {
      if (class(fragments) != "matrix") {
        stop("fragments must be a matrix")
      }

      if (ncol(fragments) != 2) {
        stop("fragments must be a two-column matrix")
      }

      if (any(apply(fragments, 1, function(x) x[1]>=x[2]))) {
        stop("Invalid fragments matrix")
      }

      if (any(diff(fragments[,2]) <= 0)) {
        stop("Invalid fragments matrix")
      }
    }
    
    FUN <- function(x, filePath, info, lane, destdir, batchname, fragments, ...) {
      if (!require(srfim))
        stop("Package srfim not available")
      if (!require(SRAdata))
        stop("Package SRAdata not available")

      tab <- SRATable(filePath)
      tab@info <- info

      sumfun <- selectMethod("summary", "SrfimResult")
      
      if (is.null(fragments)) {
        filename <- sprintf("%s_%d_%04d_srfim.fq", batchname, lane, x)
        filename <- file.path(destdir, filename)
        
        res <- srfim(tab, lane=lane, tile=x, filename=filename, ...)
        sumfun(res)
      } else {
        reslist <- vector("list", nrow(fragments))
        for (i in 1:nrow(fragments)) {
          cycles <- seq(fragments[i,1], fragments[i,2])
          filename <- sprintf("%s_%d_%d_%04d_srfim.fq", batchname, lane, i, x)
          filename <- file.path(destdir, filename)

          res <- srfim(tab, lane=lane, tile=x, filename=filename, cycles=cycles, ...)
          reslist[[i]] <- sumfun(res)
        }
        return(reslist)
      }
    }

    res <- ShortRead::srapply(tiles, FUN=FUN, filePath=object@path, info=info, lane=lane, destdir=destdir,
                              batchname=object@batchName, fragments=fragments, ...)
    return(res)
  }

.srfim.SrfimBatch <- function(dirPath, pattern=character(0),
                              destdir="", ...)
                              
  {
    object <- dirPath
    if (object@type == "SRATable") {
      return(.srfim.SrfimBatch.SRATable(object, destdir, ...))
    }
    
    # find what to run on
    X <- switch(object@type,
                SolexaIntensity=.srfimGetFiles.SolexaIntensity(object@path, pattern=pattern,
                  intExtension="_int.txt", ...),
                Sig2Intensity=.srfimGetFiles.SolexaIntensity(object@path, pattern=pattern,
                  intExtension="_sig2.txt", ...),
                stop("File type ", object@type, " not implemented yet. Ask the maintainer to implement it!"))
    # run
    dirPath <- object@path
    preprocess <- object@type != "Sig2Intensity"
    namepat <- switch(object@type,
                      SolexaIntensity="_int.txt",
                      Sig2Intensity="_sig2.txt",
                      "")
    namesub <- "_srfim.fq"
    if ( missing(destdir) || nchar(destdir)==0)
      destdir <- getwd()
    
    FUN <- function(x, dirPath, destdir, namepat, namesub, ...) {
      if (!require(srfim))
        stop("Package srfim not available")
      filename <- file.path(destdir, gsub("\\.gz", "", gsub(namepat, namesub, x)))
      res <- srfim(dirPath, pattern=x, filename=filename, ...)
      sumfun <- selectMethod("summary", "SrfimResult")
      sumfun(res)
    }
    
    res <- ShortRead::srapply(X, FUN=FUN, dirPath=dirPath, destdir=destdir, namepat=namepat, namesub=namesub,
                              preprocess=preprocess, intExtension="", ...)
    return(res)
  }
                              
