do.linemodel <- function(u, xloc, yloc, nreads.per.block=7000,
                         probs=rep(1/4,4), fix.probs=FALSE,
                         tol=0.01,maxit=10,junk.threshold=1e-6,
                         verbose=FALSE,debug=0)
  {
    nreads <- dim(u)[1]
    ncycles <- dim(u)[3]

    X <- cbind(1,0:(ncycles-1))
    R <- chol(crossprod(X))

    z <- array(0.0,dim=dim(u))

    block.factor <- split.sample(xloc, yloc, nreads.per.block=nreads.per.block)
    blocks <- levels(block.factor)
    nblocks <- length(blocks)
    
    line.pars <- matrix(0, nr=nreads, nc=6)
    colnames(line.pars) <- c("alpha_0","alpha_1","sigma_a",
                             "beta_0","beta_1","sigma_b")
    
    sigma <- matrix(0, nr=nreads, nc=2)
    colnames(sigma) <- c("sigma_a", "sigma_b")

    mu <- list(alpha=array(0, dim=c(4, ncycles, nblocks)),
               beta=array(0, dim=c(4, ncycles, nblocks)))
    
    junk <- vector("logical", nreads)
    
    for (i in 1:nblocks)
      {
        block <- blocks[i]
        reads.in.block <- which(block.factor == block)
        
        if (verbose)
          message("Estimating parameters for block ", i, " of ", nblocks, ": ", length(reads.in.block), " reads")

        tmp <- try(linemodel(u[reads.in.block,,], X, R, ncycles,
                             probs, fix.probs,
                             tol, maxit, junk.threshold,
                             verbose, debug))

        if (is.null(tmp) | inherits(tmp,"try-error")) {
          if (verbose)
            message("Error estimating parameters of block ", i)
          
          junk[reads.in.block] <- TRUE
          z[reads.in.block,,] <- 0
          line.pars[reads.in.block,] <- 0
          mu$alpha[,,i] <- 0
          mu$beta[,,i] <- 0
        } else {
          good.reads <- reads.in.block[!tmp$filter]
          junk[reads.in.block] <- tmp$filter
          
          z[good.reads,,] <- tmp$z
          line.pars[good.reads, c("alpha_0", "alpha_1")] <- t(tmp$alpha)
          line.pars[good.reads, c("beta_0", "beta_1")] <- t(tmp$beta)
          line.pars[good.reads, c("sigma_a", "sigma_b")] <- tmp$sigma
          mu$alpha[,,i] <- tmp$mua
          mu$beta[,,i] <- tmp$mub
        }
      }

    dimnames(z) <- dimnames(u)
    list(z=z, junk=junk, line.pars=line.pars, mu=mu, block.factor=block.factor)
  }

split.sample <- function(xloc, yloc,
                         nreads.per.block=5000)
  {
    nblocks <- ceiling(length(xloc)/nreads.per.block)
    nblocks <- floor(sqrt(nblocks))

    if (nblocks == 1)
      return(factor(rep(1,length(xloc))))
    
    x <- unclass(cut(xloc, nblocks))-1
    y <- unclass(cut(yloc, nblocks))-1
      
    factor(x*nblocks + y)
  }

fit <- function(yy,X,R) {
      backsolve(R,forwardsolve(R,t(yy %*% X),upper=TRUE,trans=TRUE))
    }

get.weights <- function(ra, rb, sigma_a, sigma_b)
  {
    densb <- dnorm(rb,sd=sqrt(sigma_b),log=TRUE)
    ll <- basesum(densb)

    densa <- dnorm(ra,sd=sqrt(sigma_a),log=TRUE)

    res <- densa-densb
    res <- sweep(res,c(1,3),ll,FUN="+")
    exp(res)+.Machine$double.eps
  }


get.mu <- function(r,z)
  {
    mu <- matrix(0,nr=dim(z)[2],nc=dim(z)[3])
    for (i in 1:nrow(mu))
      for (j in 1:ncol(mu)) {
        w <- z[,i,j]
        w <- w/sum(w)
        mu[i,j] <- aroma.light::weightedMedian(x=r[,i,j],w=w)
      }
    mu
  }

debug.plot <- function(y,z,alpha,beta,mua,mub,ncycles)
  {
    color <- c("green", "blue", "orange", "red")
    one.plot <- function(y,z,alpha,beta,mua,mub)
      {
        fa <- sweep(mua, 2, X%*%alpha, FUN="+")
        fb <- sweep(mub, 2, X%*%beta, FUN="+")
        Ey <- z*fa + (1-z)*fb
#        print(y1);print(y);print(z);print(mua);print(mub);print(r)
        
        cycle <- rep(1:ncycles,each=4)
        plot(cycle[y>9],y[y>9],col=rep(color,ncycles)[y>9],type="p",cex=1.4,xlab="cycle",ylab="h(int)")
        points(cycle[y>9],Ey[y>9],col=rep(color,ncycles)[y>9],cex=1.4,pch=19)
        abline(a=alpha[1]-alpha[2],b=alpha[2])
        abline(a=beta[1]-beta[2],b=beta[2])
      }
    
    layout(matrix(1:9,3,3))
    for (i in 1:9) {cat("Read ", i,"\n"); one.plot(y[i,,],z[i,,],alpha[,i],beta[,i],mua,mub)}
    readline("ptc")
  }

myrowSums <- function(x) .Internal(rowSums(x,nrow(x),ncol(x),FALSE))

basesum <- function(y)
  {
    nreads <- dim(y)[1]
    ncycles <- dim(y)[3]
    
    res <- matrix(0, nr=nreads, nc=ncycles)
    for (j in 1:ncycles)
      res[,j] <- base::rowSums(y[,,j])
    res
  }

linemodel <- function(y,X,R,ncycles,
                      probs=rep(1/4,4),fix.probs=FALSE,
                      tol=0.01,maxit=10,junk.threshold=1e-6,
                      verbose=FALSE,debug=0)
  {
    # negative values are my enemy
    y[y<0] <- min(y[y>0])
    
    # initialize
    # alpha to line at 5/6 quantile 
    # beta to line at median
    # split for mu computation at 2/3
#    pars <- apply(y,1,stats::quantile,p=c(1/2,2/3,5/6))
    nreads <- dim(y)[1]
    pars <- matrix(0, nr=3, nc=nreads)
    filter <- rep(FALSE, nreads)
    for (i in 1:nreads) {
      tmp <- quantile(y[i,,], probs=c(1/2, 2/3, 5/6), names=FALSE)

      if (min(diff(tmp)) <= junk.threshold) {
        filter[i] <- TRUE
      } else {
        pars[,i] <- tmp
      }
      
    }
    
    # throw junk reads out
    y <- y[!filter,,]
    pars <- pars[,!filter]

    nreads <- dim(y)[1]
    
    # finish initialization of alpha, beta
    alpha <- rbind(pars[3,],0)
    beta <- rbind(pars[1,],0)

    # initialize mua and mub
    mua <- matrix(0, nr=4, nc=ncycles)
    mub <- matrix(0, nr=4, nc=ncycles)

    # initialize sigma_alpha and sigma_beta
    sigma <- matrix(0, nr=nreads, nc=2)
    colnames(sigma) <- c("sigma_a", "sigma_b")
    
    for (i in 1:nreads) {
      yy <- y[i,,];

      r <- (yy[yy>pars[3,i]] - alpha[1,i])
      sa <- sum(r^2)/sum(yy>pars[2,i])

      r <- (yy[yy<=pars[3,i]] - beta[1,i])
      sb <- sum(r^2)/sum(yy<=pars[2,i])

      sigma[i,] <- c(sa, sb)
    }

    # initialize ra and rb since the E-step will use them below
    fa <- X%*%alpha
    ra <- sweep(y,c(3,1),fa)

    fb <- X%*%beta  
    rb <- sweep(y,c(3,1),fb)


    # allocate z
    zz <- array(0,dim=dim(y))

    # initialize convergence check
    previous.qval <- .Machine$double.xmax
    eps <- tol+1
    iter <- 0

    while (TRUE) {
      # E-step
      weights <- get.weights(ra, rb, sigma[,"sigma_a"], sigma[,"sigma_b"])
      zz <- sweep(weights,2,probs,FUN="*")
      ll <- basesum(zz)
      zz <- sweep(zz,c(1,3),ll,FUN="/")

      # get current value of the q function
      qval <- sum(zz*log(weights))
      qval <- qval + sum(sweep(zz,2,log(probs),FUN="*"))
      if (verbose) message("qval: ", -qval)

      eps <- abs(qval-previous.qval)/abs(previous.qval)
      # loop out if converged
      if (!(eps > tol & iter < maxit)) break
      iter <- iter + 1
      previous.qval <- qval

      if (verbose)
        message("Iteration: ", iter, " eps: ", eps)
      
      if (debug == 2) {
        debug.plot(y,zz,alpha,beta,mua,mub,ncycles)
      }

      # M-step 
      # alpha
      yy <- basesum(y*zz)
      alpha <- fit(yy,X,R)

      # beta
      yy <- basesum(y*(1-zz))/3
      beta <- fit(yy,X,R)

      # mu_alpha and mu_beta
      fa <- X%*%alpha
      ra <- sweep(y, c(3,1), fa)
      mua <- get.mu(ra, zz)

      fb <- X%*%beta
      rb <- sweep(y, c(3,1), fb)
      mub <- get.mu(rb, 1-zz)
      
      # sigma_a
      ra <- sweep(ra, c(2,3), mua)
      sigma[,"sigma_a"] <- base::rowSums(zz*ra^2)/ncycles

      # sigma_b

      rb <- sweep(rb, c(2,3), mub)
      sigma[,"sigma_b"] <- base::rowSums((1-zz)*rb^2)/(3*ncycles)

      # pi
      if (! fix.probs)
        probs <- apply(zz,2,mean)
    }

    list(z=zz, filter=filter,
         alpha=alpha, beta=beta,
         mua=mua, mub=mub,
         sigma=sigma)
  }
