#####
# srfimPreprocess
#
# do preprocessing on an intensity array

# .srfimPreprocess defined in preprocess.R
setGeneric("srfimPreprocess",
           function(ints, ...) standardGeneric("srfimPreprocess"),
           signature=c("ints"))
setMethod("srfimPreprocess", signature(ints="array"), .srfimPreprocess)
setMethod("srfimPreprocess", signature(ints="SolexaIntensity"),
          function(ints, ...) callGeneric(intensity(ints)[[,,,]], ...))

#####
# srfimModel
#
# estimate srfim model on an intensity array

# .srfimModel is defined in srfim.R
setGeneric("srfimModel",
           function(ints, info, ...) standardGeneric("srfimModel"),
           signature=c("ints", "info"))
setMethod("srfimModel",
          signature(ints="array", info="IntensityInfo"),
          .srfimModel)
setMethod("srfimModel",
          signature(ints="SolexaIntensity", info="IntensityInfo"),
          function(ints, info, ...) callGeneric(intensity(ints)[[,,,]], info, ...))
setMethod("srfimModel",
          signature(ints="SrfimPreprocessResult", info="IntensityInfo"),
          function(ints, info, ...) callGeneric(ints@intensity, info, ...))

# compute and/or return entropy
# compute.ENT defined in qa.R
setGeneric("getENT",
           function(model, ...) standardGeneric("getENT"),
           signature=c("model"))
setMethod("getENT",
          signature(model="SrfimModelResult"),
          function(model, ...) {
            if (!exists("ENT", envir=model@metrics, inherits=FALSE)) {
              assign("ENT", compute.ENT(model@probs[!model@junk.reads,,]), envir=model@metrics, inherits=FALSE) 
            }
            return(get("ENT", envir=model@metrics))
          })

# compute and/or return SNR
# compute.SNR defined in qa.R
setGeneric("getSNR",
           function(model, ...) standardGeneric("getSNR"),
           signature=c("model"))
setMethod("getSNR",
          signature(model="SrfimModelResult"),
          function(model, ...) {
            if (!exists("SNR", envir=model@metrics, inherits=FALSE)) {
              assign("SNR", compute.SNR(model@line.pars[!model@junk.reads,], dim(model@probs)[3]), envir=model@metrics, inherits=FALSE) 
            }
            return(get("SNR", envir=model@metrics))
          })

# compute and/or return SNR2
# compute.SNR2 defined in qa.R
setGeneric("getSNR2",
           function(model,...) standardGeneric("getSNR2"),
           signature=c("model"))
setMethod("getSNR2",
          signature(model="SrfimModelResult"),
          function(model, ...) {
            if (!exists("SNR2", envir=model@metrics, inherits=FALSE)) {
              assign("SNR2", compute.SNR2(model@line.pars[!model@junk.reads,], model@mu, dim(model@probs)[3]), envir=model@metrics, inherits=FALSE)
            }
            return(get("SNR2", envir=model@metrics))
          })

#######
# srfimQA
#
# compute QA metrics for srfim run

# .srfimQA defined in qa.R
setGeneric("srfimQA",
           function(model, ...) standardGeneric("srfimQA"),
           signature=c("model"))
setMethod("srfimQA",
          signature(model="SrfimModelResult"),
          .srfimQA)


###############
# filters

# functions defined in filter.R
setGeneric("srfimENTfilter",
           function(object, ...) standardGeneric("srfimENTfilter"),
           signature=c("object"))
setMethod("srfimENTfilter",
          signature(object="SrfimModelResult"),
          .srfimENTfilter.model)
setMethod("srfimENTfilter",
          signature(object="SrfimQAResult"),
          .srfimENTfilter.qa)


setGeneric("srfimSNRfilter",
           function(object, ...) standardGeneric("srfimSNRfilter"),
           signature=c("object"))
setMethod("srfimSNRfilter",
          signature(object="SrfimModelResult"),
          .srfimSNRfilter.model)
setMethod("srfimSNRfilter",
          signature(object="SrfimQAResult"),
          .srfimSNRfilter.qa)

#############
# basecalling
setGeneric("srfimBasecall",
           function(object, info, ...) standardGeneric("srfimBasecall"),
           signature=c("object", "info"))

# .srfimBasecall defined in basecaller.R
setMethod("srfimBasecall",
          signature(object="SrfimModelResult", info="IntensityInfo"),
          .srfimBasecall)

# write files with basecalls
setGeneric("srfimWrite",
           function(object, ...) standardGeneric("srfimWrite"),
           signature=c("object"))
setMethod("srfimWrite",
          signature(object="ShortReadQ"),
          writeFastq)

# .writeFuzzyFastq defined in basecaller.R
setMethod("srfimWrite",
          signature(object="FuzzyShortReadQ"),
          .writeFuzzyFastq)

##################
# srfim
setGeneric("srfim",
           function(dirPath,
                    pattern=character(0), ...) standardGeneric("srfim"),
           signature=c("dirPath"))

# .srfim defined in dispatch.R
setMethod("srfim",
          signature(dirPath="character"),
          .srfim.character)

setMethod("srfim",
          signature(dirPath="SolexaIntensity"),
          .srfim.SolexaIntensity)

setMethod("srfim",
          signature(dirPath="SRATable"),
          .srfim.SRATable)

# .srfim.SrfimBatch defined in dispatch.R
setMethod("srfim",
          signature(dirPath="SrfimBatch"),
          .srfim.SrfimBatch)

###########################
# summarize a srfim result object
setMethod("summary",
          signature(object="SrfimResult"),
          function(object) {
            res <- list(nreads=length(object@basecalls),
                        filter=object@filter.type,
                        nfiltered=sum(!object@filter),
                        filename=object@filename,
                        call=object@call)
            return(res)
          })

###########################
# print a srfim result object
setMethod("show",
          signature(object="SrfimResult"),
          function(object) {
            tmp <- summary(object)
            cat("class: ", class(object), "\n", sep="")
            cat("nreads: ", tmp$nreads, ", filter type: ", tmp$filter, ", reads filtered out: ", tmp$nfiltered, "\n", sep="")
          })

#######################
# SrfimBatch
#
# Create a container for a batch of solexa data
setGeneric("SrfimBatch",
           function(object, ...) standardGeneric("SrfimBatch"),
           signature=c("object"))

## an object from the ShortRead package pointing to the top of
## a Solexa run data path
setMethod("SrfimBatch", signature(object="SolexaPath"),
          function(object, type=c("SolexaIntensity", "Sig2Intensity",
                             "IparIntensity", "RtaIntensity"),
                   batchName=basename(experimentPath(object)), useSig2=TRUE, ...)
          {
            type <- match.arg(type)
            if (useSig2) {
              path <- baseCallPath(object)

              if(file.exists(path)) {
                intExtension <- "sig2.txt"
                files <- list.files(path, pattern=intExtension)

                if (length(files)>0) {
                  return(new("SrfimBatch",
                             path=path,
                             type="Sig2Intensity",
                             batchName=batchName))
                }
              }
            }
            
            new("SrfimBatch",
                path=imageAnalysisPath(object),
                type=type,
                batchName=batchName)
          })

# a character pointing to a directory with
# intensity files
setMethod("SrfimBatch", signature(object="character"),
          function(object,
                   type=c("SolexaIntensity", "Sig2Intensity",
                     "IparIntensity", "RtaIntensity"),
                   batchName=basename(object), ...) {
            new("SrfimBatch",
                path=object,
                type=match.arg(type),
                batchName=batchName)
          })

# an SRA dataset (package SRATable)
setMethod("SrfimBatch", signature(object="SRATable"),
          function(object, batchName=basename(object@filePath), ...) {
            new("SrfimBatch",
                path=object@filePath,
                type="SRATable",
                batchName=batchName)
          })

## method dispatch for srfim
#setGeneric("srfim",
#           function(object, ...) standardGeneric("srfim"),
#           signature=c("object"))
#
## a ShortRead object pointing to the top of a
## Solexa run data path
#setMethod("srfim", signature(object="SolexaPath"),
#          function(object,
#                   pattern=character(0),
#                   ...,
#                   control=srfimControl()) {
#            if (!control$ignoreSig2) {
#              # see if sig2 files exist
#              intExtension <- "_sig2.txt"
#              intPattern <- paste(pattern, intExtension, sep="")
#              fls <- try(ShortRead:::.file_names(baseCallPath(object), pattern=intPattern))
#              if (! inherits(fls, "try-error")) {
#                control$preprocess <- FALSE
#                return(callGeneric(baseCallPath(object), pattern=pattern,
#                                   intExtension="_sig2.txt", control=control,
#                                   type="SolexaIntensity", ...))
#              }
#            }
#            callGeneric(imageAnalysisPath(object), pattern=pattern,
#                          control=control, ...)
#            })
#          
#
## a directory of intensity files
#setMethod("srfim", signature(object="character"),
#          function(object,
#                   pattern=character(0),
#                   ...,
#                   control=srfimControl(),
#                   type=c("SolexaIntensity", "IparIntensity", "RtaIntensity"))
#          {
#            if (missing(type))
#              type <- "SolexaIntensity"
#            
#            .srfim_directory(object, pattern=pattern, control=control, type=type, ...)
#          })
#
## The result of reading an intensity file
## with ShortRead's readIntensities method
#setMethod("srfim", signature(object="SolexaIntensity"),
#          function(object, ...,
#                   control=srfimControl())
#          {
#            callGeneric(intensity(object)[[,,,]],
#                        pData(readInfo(object)),
#                        control=control,
#                        ...)
#          })
#
## on a nreads x 4 x ncycles array
## of intensities
#setMethod("srfim", signature(object="array"),
#          function(object,
#                   info=NULL,
#                   control=srfimControl(),
#                   ...)
#          {
#            if(missing(info) || is.null(info))
#              stop("info can't be missing")
#            .srfim(object, info, control=control, ...)           
#          })
#
#
##srfim <- function(x, ...)
##  UseMethod("srfim")
##
##srfim.array <- function(x, info=NULL, ...)
##  {
##    
##  }
##
##srfim.character <- function(x, keep.ints=FALSE, control=srfimControl(),...)
##  {
##    if(control$verbose)
##      message("Reading intensity file ", x)
##    
##    read.res <- intreader(x)
##    res <- .srfim(read.res$ints, read.res$info, control=control, ...)
##
##    res$read.res <- vector("list", 2)
##    names(res$read.res) <- names(read.res)
##
##    res$read.res$info <- read.res$info
##    res$read.res$ints <- NULL
##    
##    if (keep.ints)
##      res$read.res <- read.res$ints
##    res
##  }
##
