.srfimPreprocess <- function(ints, ..., verbose=FALSE, debug=FALSE)
  {
    # do crosstalk correction
    ctalk.res <- srfimCrosstalk(ints, ..., verbose=verbose, debug=debug)

    # do phasing correction
    phasing.res <- srfimPhasing(ctalk.res$y, ..., verbose=verbose, debug=debug)

    # done
    new("SrfimPreprocessResult",
        intensity=phasing.res$y,
        M=ctalk.res$M,
        P=phasing.res$P)
  }


srfimCrosstalk <- function(x, nsmp=3000, verbose=FALSE, debug=FALSE)
  {
    M <- ctalkGetMatrices(x, nsmp=nsmp, verbose=verbose, debug=debug)

    nreads <- dim(x)[1]
    ncycles <- dim(x)[3]
    
    y <- array(0.0, dim=dim(x), dimnames=dimnames(x))

    if (debug) {
      nplot.pts <- pmin(nreads, 1500)
      plot.pts <- sample(nreads, nplot.pts)
    }

    MM <- array(0.0, dim=c(4,4,ncycles),
                dimnames=list(dimnames(y)[[2]],
                  dimnames(y)[[2]], NULL))
    aa <- matrix(0, nr=4, nc=ncycles, dimnames=list(dimnames(y)[[2]],NULL))
    
    for (icycle in 1:ncycles) {
      if(verbose && !as.logical((icycle-1) %% 10)) 
        message("Correcting cycle ", icycle)

      yy <- ctalkCorrectCycle(x[,,icycle], M, verbose=verbose, debug=debug)
      y[,,icycle] <- yy$y
      MM[,,icycle] <- yy$M
      aa[,icycle] <- yy$a0
      
      if (debug) {
        pairs(x[plot.pts,,icycle])
        readline("ptc")
        pairs(y[plot.pts,,icycle])
      }
    }

    return(list(y=y, M=MM, a0=aa))
  }

ctalkCorrectCycle <- function(x, M, nsmp=3000, verbose=FALSE, debug=FALSE)
{
  nsmp <- pmin(nrow(x), nsmp)
  smp <- sample(nrow(x), nsmp)

  w <- ctalkGetWeight(x, M)

  MM <- M[[which.max(w)]]
  aa <- MM$a0
  MM <- MM$M

  if (debug) {
    nplot.pts <- pmin(nrow(x), 1500)
    plot.pts <- sample(nrow(x), nplot.pts)
    
    tmp <- rbind(aa, t(MM), x[plot.pts,])
    pairs(tmp, panel=function(x,y,...) {
      a0 <- c(x[1],y[1])
      A <- cbind(x[2:5], y[2:5])
      A <- sweep(A, 2, a0, FUN="+")
      points(x[-(1:5)], y[-(1:5)], ...)

      for (i in seq(len=nrow(A))) {
        b <- (A[i,2]-a0[2])/(A[i,1]-a0[1])
        abline(a=A[i,2]-b*A[i,1], b=b, col="blue", lwd=1.3)
      }
    })
    readline("ptc")
  }

  MM <- solve(MM)

  y <- sweep(x, 2, aa)
  y <- tcrossprod(y, MM)
  y <- sweep(y, 2, aa, FUN="+")

  if (debug) {
    pairs(y[plot.pts,])
    readline("ptc")
  }
  return(list(y=y, M=MM, a0=aa))
}

ctalkGetWeight <- function(x, M)
  {
    do.one.matrix <- function(MM)
      {
        a0 <- MM$a0
        A <- MM$A

        y <- sweep(x, 2, a0)
        dd <- tcrossprod(y, A)
        sx <- rowSums(y^2)
        dd <- sx - dd^2
        err <- 1/mean(sqrt(rowMins(dd)))
      }

    w <- sapply(M, do.one.matrix)
    sw <- sum(w)
    w <- w/sw
    return(w)
  }

ctalkGetMatrices <- function(x, cycleStep=7, nsmp=3000, verbose=FALSE, debug=FALSE)
  {
    ncycles <- dim(x)[3]
    cycles <- seq(3, ncycles, by=cycleStep)
    
    M <- vector("list", length=length(cycles))
    for (i in seq_along(cycles)) {
      icycle <- cycles[i]
      if (verbose)
        message("Computing crosstalk matrix from cycle ", icycle)
      M[[i]] <- ctalkGetMat4D(x[,,icycle], nsmp=nsmp, verbose=verbose, debug=debug)
    }
    names(M) <- as.character(cycles)
    return(M)
  }

ctalkGetMat4D <- function(x, nsmp=3000, verbose=FALSE, debug=FALSE)
  {
    if (length(dim(x)) != 2 & ncol(x) != 4)
      stop("Error in ctalkGetMat: matrix should be of 4 columns")

    if (!require(sfit, quietly=TRUE))
      stop("Package sfit required")

    if (debug) {
      nplotpts <- pmin(nrow(x), 1500)
      plot.pts <- sample(nrow(x), nplotpts)
      pairs(x[plot.pts,])
      readline("ptc")
    }
    
    nsmp <- pmin(nrow(x), nsmp)
    smp <- sample(nrow(x), nsmp)
    
    fit <- cfit(x[smp,], alpha=0.2, q=2, Q=98)
    A <- fit$M
    u <- rowSums(A^2)
    idx <- which.min(u)

    a0 <- A[idx,]
    A <- A[-idx,]
    u <- u[-idx]

    idx <- vector("integer", 4)
    idx[1] <- which.max(A[,1])
    idx[3] <- which.max(A[,3])

    
    idx[2] <- (1:4)[-idx[c(1,3)]][which.max(A[-idx[c(1,3)],2])]
    idx[4] <- (1:4)[-idx[c(1,3)]][which.max(A[-idx[c(1,3)],4])]
    
    #idx <- apply(A, 2, which.max)
    if (length(unique(idx)) != length(idx))
      stop("Error with crosstalk correct")
    
    A <- A[idx,]

    if (debug) {
      tmp <- rbind(a0,A,x[plot.pts,])
      pairs(tmp, panel = function(x,y,...) {
        a0 <- c(x[1],y[1])
        A <- cbind(x[2:5], y[2:5])
        points(x[-(1:5)], y[-(1:5)],...)
        for (i in seq(len=nrow(A))) {
          tmp2 <- rbind(A[i,], a0)
          lines(tmp2[,1],tmp2[,2],col="red",lwd=1.3)
        }
      })
      readline("ptc")
    }

    AA <- sweep(A,2,a0)
    u <- sqrt(rowSums(AA^2))
    AA <- sweep(AA,1,u,FUN="/")

    y <- sweep(x,2,a0)

    dd <- tcrossprod(y, AA)
    sx <- rowSums(y^2)
    dd <- sx - dd^2

    ii <- apply(dd, 1, which.min)
    dd1 <- dd[cbind(1:nrow(dd), ii)]
    dd1 <- sqrt(dd1)

    dd2 <- rowOrderStats(dd, which=2)
    dd2 <- sqrt(dd2)
    dd2 <- (dd2-dd1)/dd2

    cutoff1 <- quantile(dd1,.8)
    cutoff2 <- quantile(dd2,.2)
    keep <- (dd1<=cutoff1 & dd2>=cutoff2)

    yy <- y[keep,]
    ii <- ii[keep]

    keep2 <- apply(yy>0,1,any)
    yy <- yy[keep2,]
    ii <- ii[keep2]

    if (debug) {
      plot.pts2 <- sample(nrow(yy), 1500)

      tmp <- rbind(AA, yy[plot.pts2,])
      pairs(tmp, panel=function(x,y,...) {
        a0 <- c(0,0)
        A <- cbind(x[1:4], y[1:4])
        points(x[-(1:4)], y[-(1:4)])
        for (i in seq(len=nrow(A))) {
          abline(a=0, b=A[i,2]/A[i,1], col="red", lwd=1.3)
        }
      })
      readline("ptc")
    }
    
    M <- sapply(1:ncol(AA), function(i) {
      eigen(crossprod(yy[ii==i,]))$vectors[,1]
    })
    
    for (i in 1:ncol(M)) {
      ii <- which.max(abs(M[,i]))
      if (M[ii,i]<0) {
        M[,i] <- -M[,i]
      }
    }
    
    if (debug) {
      tmp <- rbind(AA, t(M), yy[plot.pts2,])
      pairs(tmp, panel=function(x,y,...) {
        A <- cbind(x[1:4], y[1:4])
        M <- cbind(x[5:8], y[5:8])
        points(x[-(1:8)], y[-(1:8)])
        for (i in seq(len=nrow(A))) {
          abline(a=0, b=A[i,2]/A[i,1], col="red", lwd=1.3)
          abline(a=0, b=M[i,2]/M[i,1], col="blue", lwd=1.3)
        }
      })
      readline("ptc")
    }

    if (debug) {
      MM <- solve(M)
      y <- tcrossprod(y,MM)
      pairs(y[plot.pts,])
      readline("ptc")
    }
    return(list(a0=a0, M=M, A=AA))
  }


srfimPhasing <- function(ints, nsample=10000, verbose=FALSE, debug=FALSE)
  {
    nreads <- dim(ints)[1]
    nsample <- pmin(nreads, nsample)
    smp <- sample(nreads, nsample)
    
    P <- estimatePhasing(ints[smp,,])
    y <- correctPhasing(ints, P)
    return(list(y=y, P=P))
  }

do.phase <- function(ny, maxinds, step=-1)
  {
    ncycles <- dim(ny)[3]
    cycles <- Filter(function(x) x+step >= 1 & x+step <= ncycles,
                 (1:ncycles))

    phase <- matrix(0.0, nr=4, nc=length(cycles))
    for (k in 1:4) {
      flags <- maxinds != k
      sinds <- flags[,cycles] & !flags[,cycles+step]
      binds <- flags[,cycles] & flags[,cycles+step]
      
      tmp <- ny[,k,cycles]
      tmp[!sinds] <- NA
      tmp <- matrix(tmp, nc=length(cycles))
      
      sm <- colMeans(tmp, na.rm=TRUE)
      
      tmp <- ny[,k,cycles]
      tmp[!binds] <- NA
      tmp <- matrix(tmp, nc=length(cycles))
      
      bm <- colMeans(tmp, na.rm=TRUE)
      phase[k,] <- sm-bm
    }
    phase
  }

estimatePhasing <- function(y) {
  if (!require(Matrix, quietly=TRUE))
    stop("library Matrix is required")
  
  nreads <- dim(y)[1]
  ncycles <- dim(y)[3]

  maxes <- matrix(0.0, nr=nreads, nc=ncycles)
  for (j in 1:ncycles)
    maxes[,j] <- rowMaxs(y[,,j])
  
  ny <- sweep(y, c(1,3), maxes, FUN="/")

  # filter reads
  seconds <- matrix(0, nr=nreads, nc=ncycles)
  for (j in 1:ncycles)
    seconds[,j] <- rowOrderStats(ny[,,j], which=3)

  keep <- (rowSums(seconds<0.6) >= pmin(12, ncycles))
  ny <- ny[keep,,]

  # get maxes
  maxinds <- matrix(0, nr=sum(keep), nc=ncycles)
  for (j in 1:ncycles)
    maxinds[,j] <- .Call("rowMaxPos", ny[,,j])
  
  phasing <- do.phase(ny,maxinds)
  prePhasing <- do.phase(ny,maxinds,1)

  phasingMatrices <- vector("list", 4)

  for (k in 1:4) {
    phasingMatrices[[k]] <- bandSparse(ncycles, k=-1:1, diagonals=list(phasing[k,], rep(1,ncycles), prePhasing[k,]))
  }

  phasingMatrices
}

correctPhasing <- function(y, phasingMatrices) {
  phasingIMatrices <- lapply(phasingMatrices, function(x) t(solve(x)))

  z <- array(0.0,dim=dim(y))
  for (k in 1:4) {
    z[,k,] <- as.matrix(y[,k,] %*% phasingIMatrices[[k]])
  }
  z
}
