# model-based sequence probability profiles
#
# Input:
#    ints: Array of intensities of size (nreads,4,ncycles)
#    info: data frame with of size (nreads,4) with columns (lane, tile, x, y)
#    sample.id: Id string of sample (to be written in ffq file read ids)
#    control: list of control parameters (see mbspp.control function)
#
# Output: list with components
#    z: probability profiles: array of size (nreads,4,ncycles)
#    line.pars: estimated parameters for effects regresion: matrix of size (nreads,6) with columns
#               1: alpha_0: intercept of signal line
#               2: alpha_1: slope of signal line
#               3: sigma_a: standard deviation of signal model
#               4: beta_0: intercept of noise line
#               5: beta_1: slope of noise line
#               6: sigma_b: standard deviation of noise model
#   mu: estimated base-cycle effects, array of size (4,nreads,nblocks) where nblocks is the number of
#               of blocks reads were spatially partitioned into
#   block.factor: factor of length nreads indicating which block each read belongs to in the spatial partition
#   filter.res: filtering result, list with components
#               reads: indices of filtered reads
#               SNR: SNR score of reads not in line.junk
#               cut: the cut-off used to filter reads with SNR
#               line.junk: reads thrown out while fitting line model
#   slog.x: started log transformed data: array of size (nreads,4,ncycles) (NULL unless return.slog.x==TRUE in control list)
#
.srfimModel <- function(ints, info, ...,
                           nreads.per.block=25000,
                           probs=rep(1/4,4),
                           fix.probs=FALSE,
                           em.tol=0.01,
                           em.maxit=10,
                           junk.threshold=1e-6,
                           return.slog.x=FALSE,
                           verbose=FALSE,
                           debug=FALSE)
  {
    # check input
    if (!is.array(ints))
      stop("ints must be an array")

    if (!length(dim(ints)) == 3)
      stop("ints must be a 3-dimensional array")

    if (!dim(ints)[2] == 4)
      stop("dimension 2 of ints must be of length 4")

    if (!inherits(info, "IntensityInfo"))
      stop("info must be an object of class IntensityInfo")

    info <- pData(info)
    xloc <- info$x
    yloc <- info$y
    
    nreads <- dim(ints)[1]
    if (!length(xloc) == nreads)
      stop("xloc vector must be of length equal to dim(y)[1]")

    if (!length(yloc) == nreads)
      stop("yloc vector must be of length equal to dim(y)[1]")
    
    if(verbose)
      message("Running srfim on ", nreads, " reads.")

    # the started log transform
    if (verbose)
      message("Doing started log transform")
    k <- min(ints)
    x <- if (k < 0) log2(ints - k + 2^-6) else log2(ints)
    
    # fit line models
    if (verbose)
      message("Fitting effect model parameters")
    
    em.res <- do.linemodel(x, xloc, yloc,
                           nreads.per.block,
                           probs,
                           fix.probs,
                           em.tol,
                           em.maxit,
                           junk.threshold,
                           verbose,debug)

    if (verbose)
      message(sum(em.res$junk), " reads thrown out during line modeling")

    new("SrfimModelResult",
        probs=em.res$z,
        line.pars=em.res$line.pars,
        mu=em.res$mu,
        block.factor=em.res$block.factor,
        junk.reads=em.res$junk,
        slog.x=if(return.slog.x) x else array(0,c(0,0,0)),
        shift=k)
  }


