#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <R.h>
#include <Rmath.h>
#include <R_ext/PrtUtil.h>
#include <Rinternals.h>
#include <Rdefines.h>
#include <R_ext/RConverters.h>
#include <R_ext/Rdynload.h>
#include <string.h>

/* get maximum/minimum values for each row/col */
SEXP rowMax(SEXP mat) {
  int i, j, tmpidx;
  int *itmp = INTEGER(getAttrib(mat, R_DimSymbol));
  int nrow = itmp[0], ncol=itmp[1];
  SEXP res;
  PROTECT(res = allocVector(REALSXP, nrow));
  double *dtmp = REAL(res);
  double *input=REAL(mat);

  for(i=0; i<nrow; i++) 
    dtmp[i]=input[i];

  for(j=1; j<ncol; j++) { /* loop for columns */
    tmpidx = j*nrow; 
    for(i=0; i<nrow; i++) { /* loop for rows */
      if(dtmp[i]<input[tmpidx+i]) 
	dtmp[i]=input[tmpidx+i];
    }
  }
  UNPROTECT(1);
  return(res);
}

SEXP colMax(SEXP mat) {
  int i, j, tmpidx;
  int *itmp = INTEGER(getAttrib(mat, R_DimSymbol));
  int nrow = itmp[0], ncol=itmp[1];
  SEXP res;
  PROTECT(res = allocVector(REALSXP, ncol));
  double *dtmp = REAL(res);
  double *input=REAL(mat);

  for(j=0; j<ncol; j++) { /* loop for columns */
    tmpidx = j*nrow;
    dtmp[j] = input[tmpidx];
    for(i=1; i<nrow; i++) { /* loop for rows */
      if(dtmp[j]<input[tmpidx+i])
        dtmp[j]=input[tmpidx+i];
    }
  }
  UNPROTECT(1);
  return(res);
}

/* Get the position of maximum value for each column.
   This can be used in calling bases/colors.
   Positions are integers from 0. */
SEXP colMaxPos(SEXP mat) {
  int i, j, tmpidx;
  int *itmp = INTEGER(getAttrib(mat, R_DimSymbol));
  int nrow = itmp[0], ncol=itmp[1];
  SEXP res;
  PROTECT(res = allocVector(INTSXP, ncol));
  itmp = INTEGER(res);
  double *input=REAL(mat), dtmp;

  for(j=0; j<ncol; j++) { 
    tmpidx = j*nrow;
    itmp[j] = 0;
    dtmp = input[tmpidx];
    for(i=1; i<nrow; i++) {
      if(dtmp<input[tmpidx+i]) {
        itmp[j]=i;
	dtmp = input[tmpidx+i];
      }
    }
  }
  UNPROTECT(1);
  return(res);
}

SEXP rowMaxPos(SEXP mat) {
  int i,j,tmpidx;
  int *itmp = INTEGER(getAttrib(mat, R_DimSymbol));
  int nrow = itmp[0], ncol=itmp[1];

  SEXP res;
  PROTECT(res = allocVector(INTSXP, nrow));

  itmp = INTEGER(res);
  double *input = REAL(mat);
  double *dtmp = (double *) R_alloc(nrow, sizeof(double));

  for (i=0; i<nrow; i++) {
    dtmp[i] = input[i];
    itmp[i] = 1;
  }
    
  for (j=1; j<ncol; j++) {
    tmpidx = j*nrow;
    for (i=0; i<nrow; i++) {
      if (dtmp[i] < input[tmpidx+i]) {
	dtmp[i] = input[tmpidx+i];
	itmp[i] = j+1;
      }
    }
  }
  UNPROTECT(1);
  return(res);
}

static inline char char_score(const double ratio, 
			      const int max, const int min, const int offset) {
  double dscore;
  int iscore;

  dscore = 10 * log10(ratio);
  if (R_FINITE(dscore)) {
    iscore = (int) (dscore < 0.0 ? dscore - 0.5 : dscore + 0.5);
  } else {
    iscore = (dscore == R_PosInf) ? max : min;
  }
      
  if (iscore > max) iscore = max;
  if (iscore < min) iscore = min;

  return((char) (iscore+offset));
}

static inline double entropy(const double p) {
  double res=log2(p);

  res = R_FINITE(res) ? -p * res : 0.0;
  return(res);
}


static inline char char_entropy(const double e, const double emax, const double smax, const int offset)
{
  double dres;
  int ires;

  dres = -log(e);
  //  dres = dres / (emax / smax);
  dres = (dres > smax) ? smax : dres;

  ires = (int) (dres + 0.5);

  /*  Rprintf("ent: %.2f score: %.2f smax: %.2f offset: %d ires: %d charres: %c\n",
   *	  e, dres, smax, offset, ires, (char) (ires+offset));
   */

  return ((char) (ires+offset));
}

/* call bases/colors from 3D matrix (nread x 4 x ncycle).
   This returns nread strings, each is ncycle long. 
   Also a vector of integer qualities of length nread*ncycle */
SEXP callBases(SEXP mat, SEXP type, SEXP score) {
  int iread, icolor, icycle, tmpidx, tmpidx2;
  double *input=REAL(mat), dtmp, dtmp2, dscore, ent;
  int *itmp = INTEGER(getAttrib(mat, R_DimSymbol)), itmp2;

  char call_type = *CHARACTER_VALUE(type);
  char score_type = *CHARACTER_VALUE(score);

  int nread = itmp[0], ncolor=itmp[1], ncycle=itmp[2];
  char *tmpchar = (char*) R_alloc(ncycle+1, sizeof(char));
  tmpchar[ncycle+1] = '\0';

  char *tmpscore = (char *) R_alloc(ncycle, sizeof(int));
  tmpscore[ncycle+1] = '\0';

  SEXP calls, quals, res;
  PROTECT(calls = allocVector(STRSXP, nread));
  PROTECT(quals = allocVector(STRSXP, nread));
  PROTECT(res = allocVector(VECSXP, 2));

  /* look up table */
  char color_tbl[4]={'0', '1', '2', '3'};
  char nucl_tbl[4]={'A', 'C', 'G', 'T'}; 
  char *tbl;


  /*  Rprintf("%c %c\n", call_type, score_type);*/

  tbl = (call_type == 'n') ? nucl_tbl : color_tbl;
  
  for(iread=0; iread<nread; iread++) { /* call base for this read */
    for(icycle=0; icycle<ncycle; icycle++) {
      tmpidx = icycle*nread*ncolor + iread;
      dtmp = input[tmpidx]; dtmp2=0.0; ent=entropy(dtmp);
      itmp2 = 0;
      /* compare to other colors */
      tmpidx2 = tmpidx;

      /*      Rprintf("iread %d, icycle %d, icolor 0, dtmp %.4f dtmp2 %.4f\n",iread, icycle, dtmp, dtmp2);*/
      for(icolor=1; icolor<ncolor; icolor++) {
	tmpidx2 = tmpidx2+nread;
	ent += entropy(input[tmpidx2]);
	if(dtmp < input[tmpidx2]) {
	  dtmp2 = dtmp;
	  dtmp = input[tmpidx2]; 
	  itmp2 = icolor;
	} else {
	  if (dtmp2 < input[tmpidx2])
	    dtmp2 = input[tmpidx2];
	}
	/*	Rprintf("iread %d, icycle %d, icolor %d, p %.4f dtmp %.4f dtmp2 %.4f\n",iread, icycle, icolor, input[tmpidx2], dtmp, dtmp2);*/
      }
      tmpchar[icycle] = tbl[itmp2];

      switch (score_type) {
      case 'p': 
	/* compute phred score */
	tmpscore[icycle] = char_score(1.0/(1.0-dtmp),40,0,33);
	break;
	
      case 'l': 
	/* compute log ratio score */
	tmpscore[icycle] = char_score(dtmp / (1.0-dtmp), 40, -5, 64);
	break;
	
      case 'm':
	/* compute log ratio score, comparing to second max prob */
	tmpscore[icycle] = char_score(dtmp / dtmp2, 40 , 0, 64);
	/*	Rprintf("iread %d icycle %d dtmp %.4f dtmp2 %.4f dscore %.4f tmpscore %d\n", iread, icycle, dtmp, dtmp2, dscore, tmpscore[icycle]);*/
	break;

      case 'e':
	/* use log of entropy as score */
	tmpscore[icycle] = char_entropy(ent, 65.0, 40.0, 33);
	break;
      }
    }

    /* assign call  */
    SET_STRING_ELT(calls, iread, mkChar(tmpchar));
    SET_STRING_ELT(quals, iread, mkChar(tmpscore));
  }

  SET_VECTOR_ELT(res, 0, calls);
  SET_VECTOR_ELT(res, 1, quals);
  UNPROTECT(3);
  return(res);
}

/* call bases/colors from 3D matrix (nread x 4 x ncycle)
   but gives alternative calls in positions with high uncertainty
   This returns nread strings, each ncycle long
   a vector of log ratios of length nread*ncycle
   a vector of alternative counts of length nread*ncycle
   nread extra strings of length equal to the sum of the number of alternatives
   a vector of more log ratios of length equal to the sum of the number of alternatives
*/
SEXP callProbs(SEXP mat, SEXP type)
{
  int iread, icolor, icolor2, icycle, tmpidx, tmpidx2;
  double *input=REAL(mat);
  int *itmp = INTEGER(getAttrib(mat, R_DimSymbol));
  int nalt_calls, inalt, inalt2;

  double dtmp[4], dscore[4], ent, p;
  int rank[4];

  char call_type= *CHARACTER_VALUE(type);

  int nread = itmp[0], ncolor=itmp[1], ncycle=itmp[2];
  char *tmpchar = (char *) R_alloc(ncycle+1, sizeof(char));
  tmpchar[ncycle+1] = '\0';

  char *tmpscore = (char *) R_alloc(ncycle+1, sizeof(char));
  char *tmpnalt = (char *) R_alloc(ncycle+1, sizeof(char));
  char *tmpalt_calls = (char *) R_alloc(3*ncycle+1, sizeof(char));
  char *tmpalt_scores = (char *) R_alloc(3*ncycle+1, sizeof(char));

  tmpscore[ncycle+1] = '\0';
  tmpnalt[ncycle+1] = '\0';
  tmpalt_calls[3*ncycle+1] = '\0';
  tmpalt_scores[3*ncycle+1] = '\0';

  SEXP calls, quals, nalt, alt_calls, alt_scores, res;
  PROTECT(calls = allocVector(STRSXP, nread));
  PROTECT(quals = allocVector(STRSXP, nread));
  PROTECT(nalt = allocVector(STRSXP, nread));
  PROTECT(alt_calls = allocVector(STRSXP, nread));
  PROTECT(alt_scores = allocVector(STRSXP, nread));

  PROTECT(res = allocVector(VECSXP, 5));

  /* lookup table */
  char color_tbl[4] = {'0', '1', '2', '3'};
  char nucl_tbl[4] = {'A', 'C', 'G', 'T'};
  char *tbl;

  tbl = (call_type == 'n') ? nucl_tbl : color_tbl;

  for (iread=0; iread<nread; iread++) {
    nalt_calls=0;

    for (icycle=0; icycle<ncycle; icycle++) {
      tmpidx = icycle*nread*ncolor + iread;
      p = input[tmpidx];

      rank[0] = 0; dtmp[0] = p;
      ent = entropy(p);
      inalt = 0; 
      
      #ifdef HCB_DEBUG 
      Rprintf("read %d, cycle %d, p[0]=%.4f, ent=%.4f, nalt=%d\n",iread,icycle,p,ent,inalt);
      #endif

      tmpidx2=tmpidx;
      for (icolor=1; icolor<4; icolor++) {
	tmpidx2=tmpidx2+nread;
	p = input[tmpidx2];

	ent += entropy(p);
	if (ent > log2(inalt+1.3)) inalt++;
	
	#ifdef HCB_DEBUG
	Rprintf("p[%d]=%.4f, ent=%.4f, nalt=%d\n", icolor, p, ent, inalt);
	#endif

	for (icolor2=0; icolor2<icolor; icolor2++) {
	  if (p >= dtmp[icolor2]) {
	    memmove(rank+icolor2+1,rank+icolor2,sizeof(int) * (3-icolor2));
	    memmove(dtmp+icolor2+1,dtmp+icolor2,sizeof(double) * (3-icolor2));
	    break;
	  }
	}
	rank[icolor2]=icolor;
	dtmp[icolor2]=p;

	#ifdef HCB_DEBUG
	Rprintf("rank: %d", rank[0]);
	for (icolor2=1;icolor2<=icolor;icolor2++)
	  Rprintf(", %d", rank[icolor2]);
	Rprintf("\n");

	Rprintf("dtmp: %.4f", dtmp[0]);
	for (icolor2=1;icolor2<=icolor;icolor2++)
	  Rprintf(", %.4f", dtmp[icolor2]);
	Rprintf("\n");
	#endif
      }

      tmpchar[icycle] = tbl[rank[0]];
      tmpscore[icycle] = char_score(dtmp[0] / dtmp[1], 40 ,-30, 64);

      tmpnalt[icycle] = (char) (48+inalt);

      for (inalt2=0; inalt2<inalt; inalt2++) {
	tmpalt_calls[nalt_calls] = tbl[rank[inalt2+1]];
	tmpalt_scores[nalt_calls] = char_score(dtmp[inalt2+1]/dtmp[0], 40, -30, 64);

	#ifdef HCB_DEBUG
	Rprintf("inalt2 %d inalt %d call %c log_ratio %.4f score %c\n",
	  inalt2, inalt, tmpalt_calls[nalt_calls], 10*log10(dtmp[inalt2+1]/dtmp[0]), tmpalt_scores[nalt_calls]); 
	#endif

	nalt_calls++;
      }
    }

    tmpalt_calls[nalt_calls] = '\0';
    tmpalt_scores[nalt_calls] = '\0';

    SET_STRING_ELT(calls, iread, mkChar(tmpchar));
    SET_STRING_ELT(quals, iread, mkChar(tmpscore));
    SET_STRING_ELT(nalt, iread, mkChar(tmpnalt));
    SET_STRING_ELT(alt_calls, iread, mkChar(tmpalt_calls));
    SET_STRING_ELT(alt_scores, iread, mkChar(tmpalt_scores));
  }

  SET_VECTOR_ELT(res, 0, calls);
  SET_VECTOR_ELT(res, 1, quals);
  SET_VECTOR_ELT(res, 2, nalt);
  SET_VECTOR_ELT(res, 3, alt_calls);
  SET_VECTOR_ELT(res, 4, alt_scores);

  UNPROTECT(6);
  return(res);
}
